<?php
class string_analytics {
	// 分詞的來源字串
	private $sourceString;
	// 初步分析的結果
	private $simpleResult;
	// 分詞的結果
	private $finallyResult;
	// 字典檔
	private $dict;
	// 文字解析方向
	private $gran_type;
	// 詞彙最大長度
	private $max_words;
	// 反回標籤數量
	private $max_tags;

	function __construct() {
		$this->sourceString = '';
		$this->simpleResult = array();
		$this->n_gramResult = array();
		$this->finallyResult = array();
		$this->dict = array();
	}

	public function get_finally() {
		$return = array();
		$i = $this->max_tags;
		foreach( $this->finallyResult as $key => $times ) {
			$return[] = htmlspecialchars($key);
			$i--;
			if( $i <= 0 ) {
				break;
			}
		}
		return $return;
	}

	public function getSourceString() {
		return htmlspecialchars($this->sourceString);
	}

	public function SetOption($type, $max_words, $max_tags) {
		$this->gran_type = $type;
		$this->max_words = $max_words;
		$this->max_tags = $max_tags;
	}

	public function SetDict($dict, $diff = 1) {
		foreach( $dict as $key => $value ) {
			if( isset($this->dict[$key]) ) {
				$this->dict[$key] += $value * $diff;
			} else {
				$this->dict[$key] = $value * $diff;
			}
		}
	}

	public function SetSource($source) {
		$this->sourceString = $source;
		$this->StringClear();
		$this->htmltag_overwrite();
		$this->PunctuationClear();
	}

	public function StartAnalysis() {
		$this->simpleResult = array_filter(explode(' ', $this->sourceString));
		$this->sourceString = implode(' ', $this->simpleResult);
		if( $this->gran_type == 'f' ) {
			$this->forward_n_gram();
		} else {
			$this->reverse_n_gram();
		}
		$this->makefinally();
	}

	private function StringClear() {
		// NCR 轉換回字符
		$this->sourceString = preg_replace_callback('/&#([0-9]{1,6}|x[0-9A-F]{1,5});/i', array(&$this, 'decode_ncr_callback'), $this->sourceString);
		// html實體字符轉換回字符
		$this->sourceString = htmlspecialchars_decode($this->sourceString, ENT_QUOTES);
		// 移除特殊字符 
		$sb_match = "\r\n\t\x00\x01\x02\x03\x04\x05\x06\x07\x08\x0B\x0C\x0E\x0F\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1A\x1B\x1C\x1D\x1E\x1F\xC0\xC1\xF5\xF6\xF7\xF8\xF9\xFA\xFB\xFC\xFD\xFE\xFF";
		$sb_replace	= '                                              ';
		$this->sourceString = strtr($this->sourceString, $sb_match, $sb_replace);
		// 全型轉半型 ( 0xFF00 ~ 0xFF5F => 0x20 ~ 0x7F )
		$sb_match = array('　', '＀', '！', '＂', '＃', '＄', '％', '＆', '＇', '（', '）', '＊', '＋', '，', '－', '．', '／', '０', '１', '２', '３', '４', '５', '６', '７', '８', '９', '：', '；', '＜', '＝', '＞', '？', '＠', 'Ａ', 'Ｂ', 'Ｃ', 'Ｄ', 'Ｅ', 'Ｆ', 'Ｇ', 'Ｈ', 'Ｉ', 'Ｊ', 'Ｋ', 'Ｌ', 'Ｍ', 'Ｎ', 'Ｏ', 'Ｐ', 'Ｑ', 'Ｒ', 'Ｓ', 'Ｔ', 'Ｕ', 'Ｖ', 'Ｗ', 'Ｘ', 'Ｙ', 'Ｚ', '［', '＼', '］', '＾', '＿', '｀', 'ａ', 'ｂ', 'ｃ', 'ｄ', 'ｅ', 'ｆ', 'ｇ', 'ｈ', 'ｉ', 'ｊ', 'ｋ', 'ｌ', 'ｍ', 'ｎ', 'ｏ', 'ｐ', 'ｑ', 'ｒ', 'ｓ', 'ｔ', 'ｕ', 'ｖ', 'ｗ', 'ｘ', 'ｙ', 'ｚ', '｛', '｜', '｝', '～');
		$sb_replace = array(' ', ' ', '!', '"', '#', '$', '%', '&', '\'', '(', ')', '*', '+', ',', '-', '.', '/', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', ':', ';', '<', '=', '>', '?', '@', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '[', '\\', ']', '^', '_', '`', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', '{', '|', '}', '~');
		$this->sourceString = str_replace($sb_match, $sb_replace, $this->sourceString);
	}

	private function decode_ncr_callback($matches) {
		$cp = (strncasecmp($matches[1], 'x', 1)) ? $matches[1] : hexdec(substr($matches[1], 1));
		$cp = $matches;
		if( $cp > 0xFFFF ) {
			return chr(0xF0 | ($cp >> 18)) . chr(0x80 | (($cp >> 12) & 0x3F)) . chr(0x80 | (($cp >> 6) & 0x3F)) . chr(0x80 | ($cp & 0x3F));
		} elseif( $cp > 0x7FF ) {
			return chr(0xE0 | ($cp >> 12)) . chr(0x80 | (($cp >> 6) & 0x3F)) . chr(0x80 | ($cp & 0x3F));
		} elseif( $cp > 0x7F ) {
			return chr(0xC0 | ($cp >> 6)) . chr(0x80 | ($cp & 0x3F));
		} else {
			return chr($cp);
		}
	}

	private function htmltag_overwrite() {
		preg_match_all('/<h[1-3][^>]*>([^<]*)<\/h[1-3]\s*>/s', $this->sourceString, $match);
		if( count($match[1]) > 0 ) {
			$this->sourceString .= ' ' . implode(' ', $match[1]) .  ' ' . implode(' ', $match[1]);
		}
		preg_match_all('/<h[4-6][^>]*>([^<]*)<\/h[4-6]\s*>/s', $this->sourceString, $match);
		if( count($match[1]) > 0 ) {
			$this->sourceString .= ' ' . implode(' ', $match[1]);
		}
		while( preg_match_all('/<[b|i|em|strong|span][^>]*>([^<]*)<\/[b|i|em|strong|span]\s*>/s', $this->sourceString, $match) > 0 ) {
			$this->sourceString = str_replace($match[0], $match[1], $this->sourceString);
			$this->sourceString .= ' ' . implode(' ', $match[1]);
		}
		preg_match_all('/alt="([^\"]*)"/s', $this->sourceString, $match);
		if( count($match[1]) > 0 ) {
			$this->sourceString .= ' ' . implode(' ', $match[1]);
		}
		// 移除 html 標籤
		$this->sourceString = strip_tags($this->sourceString);
	}

	private function PunctuationClear() {
		// 英文轉小寫 標點符號空白
		$sb_match = 'ISTCPAMELRDOJBNHFGVWUQKYXZ!"#$%&\'()*+,/:;<=>?@[\\]^`{|}~';
		$sb_replace	= 'istcpamelrdojbnhfgvwuqkyxz                             ';
		$this->sourceString = strtr($this->sourceString, $sb_match, $sb_replace);
		// 標點符號與特殊字 轉 空白
		$sb_match = array('❤', 'ღ', '❥', '♡', '♥', '★', '✰', '☆', '✩', '✫', '✬', '✭', '✮', '✡', '', '↑', '↓', '←', '→', '↖', '↗', '↙', '↘', '↔', '↕', '➻', '➼', '➽', '➸', '➳', '➺', '➻', '➴', '➵', '➶', '➷', '➹', '▶', '➩', '➪', '➫', '➬', '➭', '➮', '➯', '➱', '➲', '➾', '➔', '➘', '➙', '➚', '➛', '➜', '➝', '➞', '➟', '➠', '➡', '➢', '➣', '➤', '➥', '➦', '➧', '➨', '↚', '↛', '↜', '↝', '↞', '↟', '↠', '↠', '↡', '↢', '↣', '↤', '↤', '↥', '↦', '↧', '↨', '⇄', '⇅', '⇆', '⇇', '⇈', '⇉', '⇊', '⇋', '⇌', '⇍', '⇎', '⇏', '⇐', '⇑', '⇒', '⇓', '⇔', '⇖', '⇗', '⇘', '⇙', '⇜', '↩', '↪', '↫', '↬', '↭', '↮', '↯', '↰', '↱', '↲', '↳', '↴', '↵', '↶', '↷', '↸', '↹', '↺', '↻', '↼', '↽', '↾', '↿', '⇀', '⇁', '⇂', '⇃', '⇞', '⇟', '⇠', '⇡', '⇢', '⇣', '⇤', '⇥', '⇦', '⇧', '⇨', '⇩', '⇪', '♞', '♘', '♙', '♟', '♕', '♛', '☬', '♝', '☫', '♚', '☃', '☸', '♗', '〠', '☊', '☋', '❡', '๑', '♨', '❖', '☇', '☈', '۞', '۩', '♤', '♠', '♧', '♣', '◇', '◆', '▧', '◘', '▩', '▣', '◙', '▨', '▤', '▥', '▦', '✠', '☜', '☞', '☎', '☏', '♂', '♀', '☼', '♈', '☄', 'ﺴ', '◈', '♦', '◊', '₪', '✌', '☃', 'ↂ', '☁', '☂', '¤', '☽', '☾', '✙', '✈', '✉', '✁', '㍿', '№', '✎', '✐', '◎', '○', '●', '⊕', '⊙', '◕', '◑', '◐', '™', '℗', 'ஐ', '℡', '✔', '☑', '☒', '☢', '▪', '▫', '□', '■', '❈', '❉', '❧', '❦', '❁', '❀', '✳', '✲', '✱', '❈', '❃', '❂', '❁', '❀', '✿', '✾', '✽', '✼', '✻', '✺', '✹', '✸', '✷', '✶', '✵', '✴', '❄', '❅', '❆', '❇', '┋', '┇', '┅', '﹉', '﹊', '﹍', '﹍', '∈', '∋', 'Π', '▬', '☍', '△', '▽', '▲', '▼', '►', '◄', '卍', '卐', '〄', '☌', '❣', '✆', '✇', '✑', '✒', '✓', '✕', '✖', '✢', '✣', '✤', '✥', '✟', '✧', '❍', '❏', '❐', '❒', '✗', '✘', '✚', '✜', '✝', '┬', '┴', '├', '─', '┼', '┤', '┌', '┐', '╞', '═', '╪', '╡', '│', '▕', '└', '┘', '╭', '╮', '╰', '╯', '╔', '╦', '╗', '╠', '═', '╬', '╣', '╓', '╥', '╖', '╒', '╤', '╕', '║', '╚', '╩', '╝', '╟', '╫', '╢', '╙', '╨', '╜', '╞', '╪', '╡', '╘', '╧', '╛', '', '＿', 'ˍ', '▁', '▂', '▃', '▄', '▅', '▆', '▇', '█', '▏', '▎', '▍', '▌', '▋', '▊', '▉', '◢', '◣', '◥', '◤', '▓', '░', '▒', '', '①', '②', '③', '④', '⑤', '⑥', '⑦', '⑧', '⑨', '⑩', '⑪', '⑫', '⑬', '⑭', '⑮', '⑯', '⑰', '⑱', '⑲', '⑳', '⓪', '', '❶', '❷', '❸', '❹', '❺', '❻', '❼', '❽', '❾', '❿', '⑴', '⑵', '⑶', '⑷', '⑸', '⑹', '⑺', '⑻', '⑼', '⑽', '㈠', '㈡', '㈢', '㈣', '㈤', '㈥', '㈦', '㈧', '㈨', '㈩', '㊀', '㊁', '㊂', '㊃', '㊄', '㊅', '㊆', '㊇', '㊈', '㊉', '１', '２', '３', '４', '５', '６', '７', '８', '９', '０', 'Ⅰ', 'Ⅱ', 'Ⅲ', 'Ⅳ', 'Ⅴ', 'Ⅵ', 'Ⅶ', 'Ⅷ', 'Ⅸ', 'Ⅹ', 'Ⅺ', 'Ⅻ', 'ⅰ', 'ⅱ', 'ⅲ', 'ⅳ', 'ⅴ', 'ⅵ', 'ⅶ', 'ⅷ', 'ⅸ', 'ⅹ', 'ⅺ', 'ⅻ', '', 'ⓐ', 'ⓑ', 'ⓒ', 'ⓓ', 'ⓔ', 'ⓕ', 'ⓖ', 'ⓗ', 'ⓘ', 'ⓙ', 'ⓚ', 'ⓛ', 'ⓜ', 'ⓝ', 'ⓞ', 'ⓟ', 'ⓠ', 'ⓡ', 'ⓢ', 'ⓣ', 'ⓤ', 'ⓥ', 'ⓦ', 'ⓧ', 'ⓨ', 'ⓩ', '', 'Ⓐ', 'Ⓑ', 'Ⓒ', 'Ⓓ', 'Ⓔ', 'Ⓕ', 'Ⓖ', 'Ⓗ', 'Ⓘ', 'Ⓙ', 'Ⓚ', 'Ⓛ', 'Ⓜ', 'Ⓝ', 'Ⓞ', 'Ⓟ', 'Ⓠ', 'Ⓡ', 'Ⓢ', 'Ⓣ', 'Ⓤ', 'Ⓥ', 'Ⓦ', 'Ⓧ', 'Ⓨ', 'Ⓩ', '', '⒜', '⒝', '⒞', '⒟', '⒠', '⒡', '⒢', '⒣', '⒤', '⒥', '⒦', '⒧', '⒨', '⒩', '⒪', '⒫', '⒬', '⒭', '⒮', '⒯', '⒰', '⒱', '⒲', '⒳', '⒴', '⒵ ', '￥', '￡', '€', '￠', '¥', '；', '﹔', '︰', '﹕', '：', '，', '﹐', '、', '．', '﹒', '˙', '·', '。', '？', '！', '～', '‥', '‧', '′', '〃', '〝', '〞', '‵', '‘', '’', '『', '』', '「', '」', '“', '”', '…', '❞', '❝', '﹁', '﹂', '﹃', '﹄ ', '〔', '〕', '【', '】', '﹝', '﹞', '〈', '〉', '﹙', '﹚', '《', '》', '（', '）', '｛', '｝', '﹛', '﹜', '︵', '︶', '︷', '︸', '︹', '︺', '︻', '︼', '︽', '︾', '︿', '﹀', '＜', '＞', '∩', '∪ ', '☺', '☻', '☹', '〠', '', '♫', '♬', '♪', '♩', '♭', '♪', '♮', '♯', '㊬', '㊯', '㊊', '㊋', '㊌', '㊍', '㊎', '㊏', '㊐', '㊣', '㊚', '㊛', '㊙', '㊟', '㊕', '㊗', '㊡', '㊝', '㊭', '㊮', '㊑', '㊒', '㊓', '㊔', '㊕', '㊖', '㊘', '㊜', '㊞', '㊠', '㊢', '㊤', '㊥', '㊦', '㊧', '㊨', '㊩', '㊪', '㊫', '㊰', '≒', '≌', '∵', '∴', '×', '／', '﹣', '±', '≦', '≧', '﹤', '﹥', '≠', '≡', '¼', '½', '¾', '³', '²', '∞', '√', '㏒', '∷', '°', '÷', 'ˇ', '㏑', '∫', '∮', '∠', '∟', '⊿', '＋', '﹢', '⊥', '╳', '⊾', '℅', '㎎', '㎏', '㎜', '㎝', '㎞', '㎡', '㏄', '㏎', '㏕', '℃', '℉', '‰');
		$this->sourceString = str_replace($sb_match, ' ', $this->sourceString);
	}

	private function forward_n_gram() {
		reset($this->simpleResult);
		do {
			$string = current($this->simpleResult);
			$words = array();
			if( mb_strwidth($string) == mb_strlen($string) ) {
				$words[] = $string;
				while( $string = next($this->simpleResult) ) {
					if( mb_strwidth($string) == mb_strlen($string) ) {
						$words[] = $string;
					} else {
						break;
					}
				}
				$this->forward_n_gram_make_words($words, ' ', true);
				prev($this->simpleResult);
			} else {
				$len = mb_strlen($string);
				for( $i = 0; $i < $len; $i++ ) {
					$word = mb_substr($string, $i, 1);
					if( mb_strwidth($word) == mb_strlen($word) ) {
						if( is_string($words) ) {
							$words .= $word;
						} else {
							$this->forward_n_gram_make_words($words);
							$words = $word;
						}
					} else {
						if( is_array($words) ) {
							$words[] = $word;
						} else {
							$this->forward_n_gram_make_words($words, ' ', true);
							$words = array($word);
						}
					}
				}
				$this->forward_n_gram_make_words($words);
			}
		} while( next($this->simpleResult) );
		$this->n_gramResult = array_count_values($this->n_gramResult);
	}

	private function reverse_n_gram() {
		reset($this->simpleResult);
		do {
			$string = current($this->simpleResult);
			$words = array();
			if( mb_strwidth($string) == mb_strlen($string) ) {
				$words[] = $string;
				while( $string = next($this->simpleResult) ) {
					if( mb_strwidth($string) == mb_strlen($string) ) {
						array_unshift($words, $string);
					} else {
						break;
					}
				}
				$this->reverse_n_gram_make_words($words, ' ', true);
				prev($this->simpleResult);
			} else {
				$len = mb_strlen($string);
				for( $i = $len - 1; $i >= 0; $i-- ) {
					$word = mb_substr($string, $i, 1);
					if( mb_strwidth($word) == mb_strlen($word) ) {
						if( is_string($words) ) {
							$words = $word . $words;
						} else {
							$this->reverse_n_gram_make_words($words);
							$words = $word;
						}
					} else {
						if( is_array($words) ) {
							$words[] = $word;
						} else {
							$this->reverse_n_gram_make_words($words, ' ', true);
							$words = array($word);
						}
					}
				}
				$this->reverse_n_gram_make_words($words);
			}
		} while( next($this->simpleResult) );
		$this->n_gramResult = array_count_values($this->n_gramResult);
	}

	private function forward_n_gram_make_words($words, $word_link = '', $is_en = false) {
		if( is_string($words) ) {
			if( !preg_match('/^[_\.\-]$/i', $words) ) {
				$this->n_gramResult[] = $words;
			}
			return true;
		}
		$len = count($words);
		if( $len == 1 ) {
			$temp = reset($words);
			if( !preg_match('/^[_\.\-]$/i', $temp) ) {
				$this->n_gramResult[] = $temp;
			}
			return true;
		}

		$cat = true;
		$max_len = $is_en ? 2 : $this->max_words;
		while( $len > 1 ) {
			for( $j = min($len, $max_len); $j > 1; $j-- ) {
				$temp = reset($words);
				for( $k = $j; $k > 1; $k-- ) {
					$temp .= $word_link . next($words);
				}
				if( !preg_match('/[_\.\-]/i', $temp) ) {
					$this->n_gramResult[] = $temp;
					if( isset($this->dict[$temp]) ) {
						$cat = false;
						while( $j >= 1 ) {
							array_shift($words);
							$j--;
							$len--;
						}
						break;
					}
				}
			}
			if( $cat ) {
				array_shift($words);
				$len--;
			} else {
				$cat = true;
			}
		}
	}

	private function reverse_n_gram_make_words($words, $word_link = '', $is_en = false) {
		if( is_string($words) ) {
			if( !preg_match('/^[_\.\-]$/i', $words) ) {
				$this->n_gramResult[] = $words;
			}
			return true;
		}
		$len = count($words);
		if( $len == 1 ) {
			$temp = reset($words);
			if( !preg_match('/^[_\.\-]$/i', $temp) ) {
				$this->n_gramResult[] = $temp;
			}
			return true;
		}

		$cat = true;
		$max_len = $is_en ? 2 : $this->max_words;
		while( $len > 1 ) {
			for( $start_len = $j = min($len, $max_len); $j > 1; $j-- ) {
				$temp = reset($words);
				for( $k = $j; $k > 1; $k-- ) {
					$temp = next($words) . $word_link . $temp;
				}
				if( !preg_match('/[_\.\-]/i', $temp) ) {
					$this->n_gramResult[] = $temp;
					if( isset($this->dict[$temp]) ) {
						while( $start_len >= $j ) {
							array_pop($this->n_gramResult);
							$start_len--;
						}
						$this->n_gramResult[] = $temp;
						$cat = false;
						while( $j >= 1 ) {
							array_shift($words);
							$j--;
							$len--;
						}
						break;
					}
				}
			}
			if( $cat ) {
				array_shift($words);
				$len--;
			} else {
				$cat = true;
			}
		}
	}

	private function makefinally() {
		$this->finallyResult = $this->n_gramResult;
		$this->simpleResult = array_count_values($this->simpleResult);
		$log_base = log(count($this->simpleResult) + 1);
		$dict_sum = log(array_sum($this->dict), $log_base);
		$ngram_sum = log(array_sum($this->n_gramResult), $log_base);
		$simple_sum = log(array_sum($this->simpleResult), $log_base);
		foreach( $this->finallyResult as $key => $value ) {
			$this->finallyResult[$key] = $value / $ngram_sum;
			if( isset($this->dict[$key]) ) {
				$this->finallyResult[$key] += $this->dict[$key] / $dict_sum;
			}
			if( isset($this->simpleResult[$key]) ) {
				$this->finallyResult[$key] += $this->simpleResult[$key] / $simple_sum;
			}
		}
		$max_words_result = array();
		$delete_diff = array_sum($this->finallyResult) / count($this->finallyResult);
		foreach( $this->finallyResult as $key => $value ) {
			$test = mb_substr($key, 1);
			if( isset($this->finallyResult[$test]) ) {
				if( $value == $this->finallyResult[$test] || $value - $this->finallyResult[$test] > $delete_diff ) {
					unset($this->finallyResult[$test]);
				}
				if( $this->finallyResult[$test] - $value > $delete_diff ) {
					unset($this->finallyResult[$key]);
				}
			}
			$test = mb_substr($key, 0, -1);
			if( isset($this->finallyResult[$test]) ) {
				if( $value == $this->finallyResult[$test] || $value - $this->finallyResult[$test] > $delete_diff ) {
					unset($this->finallyResult[$test]);
				}
				if( $this->finallyResult[$test] - $value > $delete_diff ) {
					unset($this->finallyResult[$key]);
				}
			}
			if( mb_strlen($key) == $this->max_words ) {
				$max_words_result[] = $key;
			}
		}
		$len = count($max_words_result) - 1;
		for( $i = 0; $i < $len; $i++ ) {
			for( $j = 1; $j < $this->max_words; $j++ ) {
				$check = mb_substr($max_words_result[$i + 1], 0, $j) . $max_words_result[$i];
				if( mb_substr_count($this->sourceString, $check) > 0 ) {
					unset($this->finallyResult[$max_words_result[$i]]);
					unset($this->finallyResult[$max_words_result[$i + 1]]);
					break;
				}
			}
		}
		arsort($this->finallyResult);
	}
}
?>